#!/usr/bin/python2.7
# -*- coding: utf-8 -*-
#
#  OX for UCS
#  listener module: handling contexts
#
# Copyright (C) 2004-2020 Univention GmbH <http://www.univention.de/>
# and
#     iKu Systems & Services GmbH & Co. KG (http://www.iku-systems.de/)
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

__package__ = ''  # workaround for PEP 366
import os
import random
import stat
import string

import listener
from univention.ox.listener_tools import (
    AsRoot,
    ListenerQueue,
    OXGroupCache,
    default_context,
    exec_cmd,
    get_local_timezone,
    get_master_credentials,
    listener_is_deactivated,
    ox_call,
    ox_list_context,
)
from univention.ox.oxlistener import OxListener
from univention.ox.oxlogger import OXLogger

try:
	from typing import List, Optional, Union
except ImportError:
	pass


name = 'ox-context'
description = 'Update Open-Xchange context'
filter = '(objectClass=oxContext)'
attributes = []

CSV_CONTEXT_COL_ID = 'id'
CSV_CONTEXT_COL_FILESTORE = 'filestore_id'

logger = None  # type: OXLogger


def create_passwd(length=20, specials='@-_+=:,.;'):  # type: (Optional[int], Optional[str]) -> str
	if not specials:
		specials = ''
	pw = list()
	if length >= 4:
		pw.append(random.choice(string.lowercase))
		pw.append(random.choice(string.uppercase))
		pw.append(random.choice(string.digits))
		if specials:
			pw.append(random.choice(specials))
		length -= len(pw)
	pw.extend(random.choice(string.ascii_letters + string.digits + specials) for _x in range(length))
	random.shuffle(pw)
	return ''.join(pw)


def get_permissions_parameter(context_id):  # type: (Union[str,int]) -> List[str]
	ox_master_admin, ox_master_pass = get_master_credentials()
	output = exec_cmd(
		[
			'/opt/open-xchange/sbin/getmoduleaccessforcontext',
			'--adminuser', ox_master_admin,
			'--adminpass={}'.format(ox_master_pass),
			'--contextid', str(context_id),
		],
		logger=logger
	)
	keys, values = output.split()
	keys = [k.strip() for k in keys.split(',')]
	values = ['on' if v == '"true"' else 'off' for v in values.split(',')]
	return ['--{}={}'.format(*kv) for kv in zip(keys, values)]


def create_context(new):
	ox_master_admin, ox_master_pass = get_master_credentials()
	context_id = new['oxContextIDNum'][0]
	context_name = 'context{}'.format(context_id)
	context_admin = 'oxadmin-{}'.format(context_name)
	context_admin_mail = '{}@{}'.format(context_admin, listener.configRegistry['domainname'])
	quota = new.get('oxQuota', ['-1'])[0]
	if quota == '0':
		quota = '-1'
	password = create_passwd()
	with AsRoot():
		with open('/etc/ox-secrets/context{}.secret'.format(context_id), 'w') as fp:
			os.fchown(fp.fileno(), 0, 0)
			os.fchmod(fp.fileno(), stat.S_IRUSR | stat.S_IWUSR)
			fp.write(password)
	_lang = listener.configRegistry.get('locale/default')
	if _lang == 'de':
		language = 'de_DE'
	else:
		language = 'en_US'
	command = [
		'createcontext',
		'--adminuser', ox_master_admin,
		'--adminpass={}'.format(ox_master_pass),
		'--contextid', context_id,
		'--contextname', context_name,
		'--addmapping', context_name,
		'--language', language,
		'--timezone', get_local_timezone(),
		'--username', context_admin,
		'--displayname', 'OX Admin',
		'--givenname', 'OX',
		'--surname', 'Admin',
		'--email', context_admin_mail,
		'--quota', quota,
		'--password', password,
	]
	command.extend(get_permissions_parameter(default_context))
	ox_call(command, logger=logger)


def modify_context(new):
	ox_master_admin, ox_master_pass = get_master_credentials()
	oxquota = new.get('oxQuota', ['-1'])[0]
	if oxquota == '0':
		oxquota = '-1'
	context_id = new['oxContextIDNum'][0]
	logger.info('setting quota: {}'.format(oxquota))

	# set quota on context
	command = [
		'changecontext',
		'--adminuser=%s' % ox_master_admin,
		'--adminpass=%s' % ox_master_pass,
		'--contextid', context_id,
		'--quota', oxquota,
	]
	ox_call(command, logger=logger)

	# get filestore id for context
	filestore_id = ''
	for row in ox_list_context(ox_master_admin, ox_master_pass):
		if row[CSV_CONTEXT_COL_ID] == new['oxContextIDNum'][0]:
			filestore_id = row[CSV_CONTEXT_COL_FILESTORE]
	# set quota on filestore
	if filestore_id:
		command = [
			'changefilestore',
			'--adminuser=%s' % ox_master_admin,
			'--adminpass=%s' % ox_master_pass,
			'--id', filestore_id,
			'--storesize', oxquota
		]
		ox_call(command, logger=logger)

	# find groups without contextid and create them in new context
	udm_groups = OxListener.get_udm_ox_groups('(!(oxContext=*))')
	context_groups = OxListener.get_ox_groups(context_id)
	for udm_group in [
		udm_group
		for udm_group in udm_groups
		if udm_group.info['name'] not in context_groups
	]:
		logger.process(
			'creating OX group "{}" in new context {}...'.format(udm_group.info['name'], context_id)
		)
		admuser, ox_master_pass = OxListener.get_credentials(context_id)

		# use description as displayname (or group name as fallback)
		group_display_name = udm_group.info.get('description', udm_group.info['name'])
		group_display_name = group_display_name[:128]  # OX length limit (Bug #35820)

		command = [
			'creategroup',
			'--displayname={}'.format(group_display_name),
			'--adminuser={}'.format(admuser),
			'--adminpass={}'.format(ox_master_pass),
			'--contextid={}'.format(context_id),
			'--name={}'.format(udm_group.info['name']),
		]
		ox_call(command, logger=logger)

		_ox_group_cache = OXGroupCache()
		_ox_group_cache.rebuild_cache_if_required()
		_ox_group_cache.add_from_ox(context_id, udm_group.info['name'], admuser, ox_master_pass)


def delete_context(old):
	# find groups with this contextid and modify them to remove it
	context_id = old['oxContextIDNum'][0]
	udm_groups = OxListener.get_udm_ox_groups('(oxContext={})'.format(context_id), modify=True)

	for udm_group in udm_groups:
		logger.process(
			'modifying UDM group "{}" to remove context {} from it...'.format(
				udm_group.info['name'], context_id
			))
		udm_group['oxContext'].remove(context_id)
		udm_group.modify()


@ListenerQueue(name, stop_on_error=False)
def handler(dn, new, old, action=None):
	global logger

	if listener_is_deactivated(name):
		return

	logger = OXLogger(name)
	if new:
		context_id = new['oxContextIDNum'][0]
	else:
		context_id = old['oxContextIDNum'][0]
	logger.qualifier = context_id

	if new and not old:
		create_context(new)
	if new and old:
		modify_context(new)
	elif old and not new:
		delete_context(old)


if __name__ == '__main__':
	handler.process_pending()
