#!/usr/bin/python2.7
# -*- coding: utf-8 -*-
#
#  OX for UCS
#  listener module: handling users
#
# Copyright (C) 2004-2020 Univention GmbH <http://www.univention.de/>
# and
#     iKu Systems & Services GmbH & Co. KG (http://www.iku-systems.de/)
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

__package__ = ''  # workaround for PEP 366
import re
import os
import datetime
from ldap.filter import filter_format

import listener
import univention.uldap
import univention.hooks
from univention.ox.oxlogger import OXLogger
from univention.ox.oxlistener import ContinueHookCommand, ExitHookCommand
from univention.ox.listener_tools import ListenerQueue, OXUserIdCache, OXGroupCache, AsRoot, get_credentials, ListenerTools_RuntimeError, ox_list_user, ox_call, listener_is_deactivated, Exclusive, get_context_admin_user


ldap2ox = {
	'oxDisplayName': 'displayname',
	'title': 'title',
	'givenName': 'givenname',
	'oxMiddleName': 'middle_name',
	'sn': 'surname',
	'oxSuffix': 'suffix',
	'oxNickName': 'nickname',
	'o': 'company',
	'oxDepartment': 'department',
	'oxPosition': 'position',
	'oxProfession': 'profession',
	'oxBranches': 'branches',
	'oxMobileBusiness': 'cellular_telephone1',
	'mobile': 'cellular_telephone2',
	'oxTelephoneCompany': 'telephone_company',
	'oxTelephoneOther': 'telephone_other',
	'oxTelephoneCar': 'telephone_car',
	'oxFaxBusiness': 'fax_business',
	'oxFaxHome': 'fax_home',
	'oxFaxOther': 'fax_other',
	'oxManagerName': 'manager_name',
	'oxTelephoneAssistant': 'telephone_assistant',
	'mailPrimaryAddress': 'email1',
	'oxEmail2': 'email2',
	'oxEmail3': 'email3',
	'oxInstantMessenger1': 'instant_messenger1',
	'oxInstantMessenger2': 'instant_messenger2',
	'oxTelephoneIp': 'telephone_ip',
	'pager': 'telephone_pager',
	'oxTelephoneTelex': 'telephone_telex',
	'oxTelephoneTtydd': 'telephone_ttytdd',
	'oxUrl': 'url',
	'oxTaxId': 'tax_id',
	'oxCommercialRegister': 'commercial_register',
	'oxSalesVolume': 'sales_volume',
	'oxMarialStatus': 'marital_status',
	'oxSpouseName': 'spouse_name',
	'street': 'street_business',
	'postalCode': 'postal_code_business',
	'l': 'city_business',
	'oxStateBusiness': 'state_business',
	'oxCountryBusiness': 'country_business',
	'oxStreetHome': 'street_home',
	'oxPostalCodeHome': 'postal_code_home',
	'oxCityHome': 'city_home',
	'oxStateHome': 'state_home',
	'oxCountryHome': 'country_home',
	'oxStreetOther': 'street_other',
	'oxPostalCodeOther': 'postal_code_other',
	'oxCityOther': 'city_other',
	'oxStateOther': 'state_other',
	'oxCountryOther': 'country_other',
	'oxUserfield01': 'userfield01',
	'oxUserfield02': 'userfield02',
	'oxUserfield03': 'userfield03',
	'oxUserfield04': 'userfield04',
	'oxUserfield05': 'userfield05',
	'oxUserfield06': 'userfield06',
	'oxUserfield07': 'userfield07',
	'oxUserfield08': 'userfield08',
	'oxUserfield09': 'userfield09',
	'oxUserfield10': 'userfield10',
	'oxUserfield11': 'userfield11',
	'oxUserfield12': 'userfield12',
	'oxUserfield13': 'userfield13',
	'oxUserfield14': 'userfield14',
	'oxUserfield15': 'userfield15',
	'oxUserfield16': 'userfield16',
	'oxUserfield17': 'userfield17',
	'oxUserfield18': 'userfield18',
	'oxUserfield19': 'userfield19',
	'oxUserfield20': 'userfield20',
	'employeeType': 'categories',
	'oxNote': 'note',
	'oxBirthday': 'birthday',
	'oxAnniversary': 'anniversary',
	'oxUserQuota': 'quota',
	}

ldap2oxmulti = {
	'telephoneNumber': 'telephone_business',
	'homePhone': 'telephone_home',
}

mapping_module_access_definitions_2_cli = {
	'calendar': 'calendar',
	'contacts': 'contacts',
	'delegatetask': 'delegate-tasks',
	'editpublicfolders': 'edit-public-folder',
	'ical': 'ical',
	'infostore': 'infostore',
	'readcreatesharedfolders': 'read-create-shared-Folders',
	'syncml': 'syncml',
	'tasks': 'tasks',
	'vcard': 'vcard',
	'webdav': 'webdav',
	'webdavxml': 'webdav-xml',
	'webmail': 'webmail',
	'editresource': 'edit-resource',
	'editgroup': 'edit-group',
	'editpassword': 'edit-password',
	'collectemailaddresses': 'collect-email-addresses',
	'multiplemailaccounts': 'multiple-mail-accounts',
	'subscription': 'subscription',
	'publication': 'publication',
	'activesync': 'active-sync',
	'deniedportal': 'denied-portal',
}


name = 'ox-user'
description = 'User settings for OX'
filter = '(&(objectClass=inetOrgPerson)(objectClass=oxUserObject))'

logger = OXLogger(name)
logger.qualifier = None

ucr_mad2cli = listener.configRegistry.get('ox/listener/user/access-definitions/mapping', '').strip()
if ucr_mad2cli:
	logger.warn('Mapping for module access definitions has been enhanced via UCR:')
	for entry in ucr_mad2cli.split():
		key, value = entry.split(':', 1)
		mapping_module_access_definitions_2_cli[key] = value
		logger.warn('  %r --> %r' % (key, value))

ucr_ldap2ox = listener.configRegistry.get('ox/listener/user/ldap/attributes/mapping/ldap2ox', '').strip()
ucr_ldap2oxmulti = listener.configRegistry.get('ox/listener/user/ldap/attributes/mapping/ldap2oxmulti', '').strip()
if ucr_ldap2ox:
	logger.warn('LDAP<->OX attribute mapping has been overwritten via UCR:')
	ldap2ox = {}
	for entry in ucr_ldap2ox.split():
		key, value = entry.split(':', 1)
		ldap2ox[key] = value
		logger.warn('  %r --> %r' % (key, value))

if ucr_ldap2oxmulti:
	logger.warn('LDAP<->OX multivalue attribute mapping has been overwritten via UCR:')
	ldap2oxmulti = {}
	for entry in ucr_ldap2oxmulti.split():
		key, value = entry.split(':', 1)
		ldap2oxmulti[key] = value
		logger.warn('  %r --> %r' % (key, value))

additional_attributes = listener.configRegistry.get('ox/listener/user/ldap/attributes', '').split()
if additional_attributes:
	logger.process('Additional LDAP attributes have been configured as trigger: %r' % (additional_attributes,))

attributes_set = set(['uid', 'uidNumber', 'gidNumber', 'mailAlternativeAddress', 'roomNumber', 'employeeNumber',
	'univentionPolicyReference', 'isOxUser', 'oxNumOfChildren',
	'oxAccess', 'oxAccessSyncMl', 'oxAccessResources', 'oxLanguage', 'oxTimeZone', 'oxAccessUSM', 'oxContextIDNum', 'oxDrive'])
attributes_set.update(ldap2ox.keys())
attributes_set.update(ldap2oxmulti.keys())
attributes_set.update(additional_attributes)
attributes = list(attributes_set)
# use the modrdn listener extension
modrdn = "1"

ldapserver = listener.configRegistry['ldap/server/name']
myhostname = listener.configRegistry['hostname']
mydomainname = listener.configRegistry['domainname']
myfqdn = '%s.%s' % (myhostname, mydomainname)
smtp_server = listener.configRegistry.get('ox/listener/smtp-server', 'localhost')
weblang = listener.configRegistry.get("locale/default", "").split(':')[0]

csv_user_row_uid = "Id"
csv_user_row_username = "Name"
csv_user_row_default_sender_address = "DefaultSenderAddress"
csv_group_row_groupname = "name"
csv_group_row_members = "members"


ACCESS_DEFINITIONS_FILE = '/opt/open-xchange/etc/ModuleAccessDefinitions.properties'
_access_definitions = dict(ts=None, defs=None)

HOOKS_DIRECTORY = '/usr/share/univention-ox/user-hooks'
hook_manager = None  # type: Union[None, univention.hooks.HookManager]

# setup cache and invalidate existing cache upon startup if not disabled otherwise
_ox_user_id_cache = OXUserIdCache(name)
_ox_group_cache = OXGroupCache(name)

try:
	with Exclusive():
		if __name__ != '__main__':   # do not rebuild cache if called by cron job
			# '__main__' ==> called by cron job
			# '/usr/lib/univention-directory-listener/system/ox-user.py' ==> called by listener
			if listener.configRegistry.is_true('ox/listener/userid-cache/autoinvalidate', True):
				_ox_user_id_cache.rebuild_cache(raise_exceptions=False)
		_ox_user_id_cache.rebuild_cache_if_required(raise_exceptions=False)
except ListenerTools_RuntimeError:
	logger.warn('could not get lock to auto-invalidate userid cache - keeping old userid cache')


def str2isodate(text):  # type: (str) -> str
	try:
		the_date = datetime.datetime.strptime(text, "%Y-%m-%d")
		return '{:%Y-%m-%d}'.format(the_date)
	except (TypeError, ValueError) as exc1:
		pass
	try:
		the_date = datetime.datetime.strptime(text, "%d.%m.%Y")
		return '{:%Y-%m-%d}'.format(the_date)
	except (TypeError, ValueError) as exc2:
		pass
	raise ValueError('Value {!r} in unknown date format or year before 1900 ({} {}).'.format(text, exc1, exc2))


def run_hook(hook_name, dn, new, old, command, *args, **kwargs):  # type: (str, str, Dict[str, List[str]], Dict[str, List[str]], str, *str, **str) -> Any
	if kwargs.get('skip_{}'.format(hook_name)):
		logger.info('Skipping hook {!r} as requested.'.format(hook_name))
		return []
	logger.info(
		'Running hook {!r} with arguments: dn={!r} new={{{} keys}} old={{{} keys}} '
		'command={!r} args={!r} kwargs={!r}'.format(
			hook_name, dn, len(new), len(old), command, args, kwargs))
	return hook_manager.call_hook(hook_name, dn, new, old, command, *args, **kwargs)


@ListenerQueue(name, stop_on_error=False)
def handler(dn, new, old, command):
	return unqueued_handler(dn, new, old, command)


def unqueued_handler(dn, new, old, command, *args, **kwargs):
	global logger, hook_manager

	if listener_is_deactivated(name):
		return

	logger = OXLogger(name)
	logger.qualifier = None

	# language
	os.environ["LANG"] = weblang

	# check if OX user id cache or OX group cache has to be rebuild
	_ox_user_id_cache.rebuild_cache_if_required()
	_ox_group_cache.rebuild_cache_if_required()

	oldIsOxUser = False
	if old:
		logger.info('Old present.')
		if old.get('isOxUser') and old['isOxUser'][0] == 'OK':
			oldIsOxUser = True

	newIsOxUser = False
	if new:
		logger.info('New present.')
		if new.get('isOxUser') and new['isOxUser'][0] == 'OK':
			newIsOxUser = True
	logger.process('New Object is OX-User: {}'.format(newIsOxUser))
	logger.process('Old Object is OX-User: {}'.format(oldIsOxUser))

	if not oldIsOxUser and not newIsOxUser:
		# Neither old nor new is OX user. Nothing to do.
		return

	blacklisted = ['Administrator', 'join-backup', 'join-slave', 'spam']

	username = ''
	if old.get('uid'):
		username = old['uid'][0]
		logger.process('old username: {}'.format(username))
	if new.get('uid'):
		username2 = new['uid'][0]
		logger.process('new username: {}'.format(username2))
		if username2 != username:
			if username2 in blacklisted or username in blacklisted:
				logger.warn('not allowed to change users in blacklist ({})'.format(username))
				return
		username = username2
	logger.qualifier = username

	if not hook_manager:
		hook_manager = univention.hooks.HookManager(HOOKS_DIRECTORY)
		logger.info('Registered hooks: {!r}'.format(hook_manager.get_hook_list()))

	groupwareoptions = []
	# some access-* options are not evaluated when creating a user.
	# track them separately and modify after user creation
	groupware_access_options = []

	if new and new.get('mailPrimaryAddress'):
			email = new['mailPrimaryAddress'][0]
			logger.process('new mailPrimaryAddress: {}'.format(email))
			groupwareoptions.append("--email=%s" % email)
			imaplogin = listener.configRegistry.get('ox/listener/imaplogin_value', '{}')
			if '{}' in imaplogin:
				imaplogin = imaplogin.format(email)
			groupwareoptions.append("--imaplogin={}".format(imaplogin))

			aliases = '%s' % email
			aliaseslist = [email]
			if new.get('mailAlternativeAddress'):
				mailalternativeaddress = new['mailAlternativeAddress']
				for alias in mailalternativeaddress:
					aliases = '%s,%s' % (aliases, alias)
					aliaseslist.append(alias)
				logger.process('aliases {}'.format(aliases))
			groupwareoptions.append("--aliases=%s" % aliases)
	elif newIsOxUser and not new.get('mailPrimaryAddress'):
		logger.error('OX-enabled user without primary email address. This is not allowed for OX users! The OX user will NOT be created, modified or deleted. Please rerun the join script to prevent this in the future:\n# univention-run-join-scripts --run-scripts --force 65univention-ox.inst')
		return
	else:
		email = ''

	oldemail = '@'
	if old and old.get('mailPrimaryAddress'):
			oldemail = old['mailPrimaryAddress'][0]
			logger.process('old mailPrimaryAddress: {}'.format(oldemail))
	#
	# Context bestimmen:
	#

	contextnum = ''
	oxhomeserver = ''
	admuser = ''
	admpass = ''
	oldcontextnum = ''
	oldoxhomeserver = ''
	oldadmuser = ''
	oldadmpass = ''
	if email:
		domain = email.rsplit('@')[-1]
	else:
		domain = oldemail.rsplit('@')[-1]
	olddomain = oldemail.rsplit('@')[-1]

	with AsRoot():
		try:
			ldapCon = univention.uldap.getMachineConnection(ldap_master=False)
		except IOError as ex:
			raise ListenerTools_RuntimeError(ex)

	if newIsOxUser:
		contextnum = get_context_id_from_user(new)
		if not contextnum:
			logger.debug('No context in NEW, using domain "{}" to find context...'.format(domain))
			contextnum = get_context_id_from_domain(ldapCon, domain)
		oxhomeserver = get_ox_home_server_for_context(ldapCon, contextnum)
		admuser, admpass = get_context_credentials(contextnum)

	if old:
		oldcontextnum = get_context_id_from_user(old)
		if not oldcontextnum:
			logger.debug('No context in OLD, using domain "{}" to find context...'.format(olddomain))
			oldcontextnum = get_context_id_from_domain(ldapCon, olddomain)
		if oldcontextnum:
			oldoxhomeserver = get_ox_home_server_for_context(ldapCon, oldcontextnum)
			oldadmuser, oldadmpass = get_context_credentials(oldcontextnum)

	# Entscheidung:
	#    old new ##
	# 1)  -   1  ## E  kein altes Objekt, lokaler Kontext: new erzeugen
	# 2)  a   1  ## E  altes Objekt nicht lokal, neues Objekt lokal: new erzeugen
	# 3)  1   -  ## L  kein neues Objekt, lokaler Kontext: old loeschen
	# 4)  1   a  ## L  altes Objekt in lokalem Kontext, neues Objekt nicht in lokalem Kontext: old loeschen
	# 5)  1   2  ## LE alter und neuer Kontext lokal: old loeschen, new erzeugen
	# 6)  1   1  ## C  altes + neues Objekt, beide lokaler Kontext: change auf new
	# 7)  -   a  ## -  kein altes Objekt, nicht lokaler Kontext: nichts
	# 8)  a   -  ## -  altes Objekt geloescht, nicht lokaler Kontext: nichts
	# 9)  a   b  ## -  alter und neuer Kontext nicht lokal: nichts

	authArgs = [
		"--adminuser=%s" % admuser,
		"--adminpass=%s" % admpass,
		"--contextid=%s" % contextnum,
	]
	oldAuthArgs = [
		"--adminuser=%s" % oldadmuser,
		"--adminpass=%s" % oldadmpass,
		"--contextid=%s" % oldcontextnum,
	]

	logger.info('email={}'.format(email))
	logger.info('domain={}'.format(domain))
	logger.info('contextnum={}'.format(contextnum))
	logger.info('oxhomeserver={}'.format(oxhomeserver))
	logger.info('oldemail={}'.format(oldemail))
	logger.info('olddomain={}'.format(olddomain))
	logger.info('oldcontextnum={}'.format(oldcontextnum))
	logger.info('oldoxhomeserver={}'.format(oldoxhomeserver))
	logger.info('myfqdn={}'.format(myfqdn))
	logger.info('admuser={}'.format(admuser))
	logger.info('oldadmuser={}'.format(oldadmuser))

	# old lokal? (3) (4) (5) (6)
	if myfqdn == oldoxhomeserver:
		logger.info('old local (3,4,5,6)')
		# context change hook
		if oldIsOxUser and newIsOxUser and contextnum != oldcontextnum:
			logger.info('Context change detected, running pre_context_changed hooks...')
			hook_results = run_hook('pre_context_changed_hook', dn, new, old, command, old_context_id=oldcontextnum, new_context_id=contextnum, *args, **kwargs)
			logger.info('Hooks returned {!r}'.format(hook_results))
			for result in hook_results:
				if isinstance(result, ContinueHookCommand):
					pass
				elif isinstance(result, ExitHookCommand):
					return result.exit_code
				elif isinstance(result, Exception):
					raise result

		if contextnum != oldcontextnum or not newIsOxUser:
			# loeschen (3) (4) (5)
			logger.info('loeschen (3) (4) (5)')

			oldoxuidnumber = _ox_user_id_cache.query_ox_id(oldcontextnum, username) or ''
			logger.process('Processing uid {} for old object'.format(oldoxuidnumber))

			for grp in _ox_group_cache.query_one_context(oldcontextnum):
				logger.debug('Processing grp: {}'.format(grp))
				if oldoxuidnumber in grp.member_list:
					if len(grp.member_list) <= 1:  # group has no members or OX user is last group member
						logger.process('removing ox group "{}" from context {}'.format(grp.groupname, oldcontextnum))
						groupdelete = [
							"deletegroup",
							"--name=%s" % (grp.groupname,),
						] + oldAuthArgs
						ox_call(groupdelete, "com.openexchange.admin.rmi.exceptions.NoSuchObjectException", logger=logger)
						_ox_group_cache.remove(oldcontextnum, grp.groupname)

			if username == oldadmuser:
				logger.info('Not deleting admin-user {}'.format(username))
			else:
				if command == 'r':
					# object has moved, write ldap uidNumber into filename and oxuidnumber in this file.
					logger.info('Command = r')
					lockfilename = '/var/lock/deleteuid-%s' % old['uidNumber'][0]
					with AsRoot():
						with open(lockfilename, 'w') as lockfilehandle:
							lockfilehandle.write(oldoxuidnumber)
					logger.process('setting lock file for user uidNumber: {}'.format(old['uidNumber'][0]))
				else:
					deleteuser = [
						"deleteuser",
						"--username=%s" % username,
					] + oldAuthArgs
					logger.process('deleting ox user "{}" from context {}'.format(username, oldcontextnum))
					ox_call(deleteuser, "com.openexchange.admin.rmi.exceptions.NoSuchObjectException", logger=logger)

					# remove OX user from user id cache and group cache
					oldoxuidnumber = _ox_user_id_cache.query_ox_id(oldcontextnum, username)
					_ox_user_id_cache.remove(oldcontextnum, username)
					if oldoxuidnumber is not None:
						for grp in _ox_group_cache.query_one_context(oldcontextnum):
							if oldoxuidnumber in grp.member_list:
								_ox_group_cache.remove_members(oldcontextnum, grp.groupname, oldoxuidnumber)

	# new lokal? (1) (2) (5) (6)
	if myfqdn == oxhomeserver:
		logger.info('new lokal (1) (2) (5) (6)')
		# ldapsearch -x gidNumber=5001
		# ldapsearch -x "(&(objectClass=posixGroup)(gidNumber=5001))" cn
		# ldapsearch -x "(&(objectClass=posixGroup)(memberUid=aaa))" cn

		if new.get('gidNumber'):
			gidnumber = new['gidNumber'][0]
		else:
			gidnumber = old['gidNumber'][0]

		filter_s = filter_format('(&(gidNumber=%s)(objectClass=posixGroup))', (str(gidnumber),))
		result_id = ldapCon.search(filter=filter_s, attr=['cn'])
		if result_id:
			prigroup = result_id[0][1].values()[0][0]
			# logger.info('prigroup={}'.format(prigroup))
		else:
			logger.error('cannot find primary group with gidNumber %r for user %r' % (gidnumber, username))
# 		filter_s = filter_format('(&(memberUid=%s)(objectClass=posixGroup))', (username,))
# 		result_id = ldapCon.search(base=ldapbase, filter=filter_s, attr=['cn'])
# 		if result_id:
# 			othergroups = result_id
# 			logger.info('othergroups={}'.format(othergroups))

		imapserver = 'localhost'
		if listener.configRegistry.get('ox/role') != 'oxae':
			filter_s = filter_format('(&(cn=%s)(objectClass=oxMailDomainServerSettings))', (domain,))
			result_id = ldapCon.search(filter=filter_s, attr=['oxImapServer'])
			if result_id:
				imapserver = result_id[0][1].values()[0][0]
			logger.info('imapserver={}'.format(imapserver))

		groupwareoptions.append("--imapserver=%s" % imapserver)
		groupwareoptions.append("--smtpserver=%s" % smtp_server)

		if new.get('oxNumOfChildren'):
			groupwareoptions.append("--number_of_children=%s" % new['oxNumOfChildren'][0])
		elif old and old.get('oxNumOfChildren'):
			groupwareoptions.append("--number_of_children=")

		if new.get('roomNumber'):
			groupwareoptions.append("--room_number=%s" % new['roomNumber'][0])
		elif old and old.get('roomNumber'):
			groupwareoptions.append("--room_number=")

		if new.get('employeeNumber'):
			groupwareoptions.append("--number_of_employee=%s" % new['employeeNumber'][0])
		elif old and old.get('employeeNumber'):
			groupwareoptions.append("--number_of_employee=")

		oxuidnumber = get_ox_id(
			admuser or oldadmuser,
			admpass or oldadmpass,
			contextnum or oldcontextnum,
			username)
		if oxuidnumber:
			logger.info('old object present (oxuidnumber={})'.format(oxuidnumber))

		if new.get('oxAccess'):
			oxaccess = new['oxAccess'][0]
			oxaccess_usm = new.get('oxAccessUSM', ['Not'])[0] == 'OK'
			ox_drive_enabled = new.get('oxDrive', ['1'])[0] == '1'
			groupwareoptions.extend(get_access_cmdline_args(oxaccess, oxuidnumber, oxaccess_usm, ox_drive_enabled))
		else:
			groupwareoptions.append("--mailenabled=false")

		if new.get('oxAccessSyncMl'):
			groupwareoptions.append("--access-syncml=on")
		else:
			groupwareoptions.append("--access-syncml=off")

		if new.get('oxAccessResources'):
			groupwareoptions.append("--access-edit-resource=on")
		else:
			groupwareoptions.append("--access-edit-resource=off")

		for ldap_attr, ox_attr in ldap2ox.items():
			if new.get(ldap_attr):
				new_ldap_value = new[ldap_attr][0]
				if ldap_attr in ('oxAnniversary', 'oxBirthday'):
					new_ldap_value = str2isodate(new_ldap_value)
				groupwareoptions.append("--%s=%s" % (ox_attr, new_ldap_value))
			elif old.get(ldap_attr):
				groupwareoptions.append("--%s=" % ox_attr)

		# if more than one value is present then set the first value into "${key}1" and the second into "${key}2"
		for ldapattr, oxmultikey in ldap2oxmulti.items():
			if new and new.get(ldapattr):
				groupwareoptions.append("--%s1=%s" % (oxmultikey, new.get(ldapattr)[0]))
				if len(new.get(ldapattr)) > 1:
					groupwareoptions.append("--%s2=%s" % (oxmultikey, new.get(ldapattr)[1]))
			elif old and old.get(ldapattr):
				groupwareoptions.append("--%s1=" % oxmultikey)
				groupwareoptions.append("--%s2=" % oxmultikey)

		groupwareoptions.append("--gui_spam_filter_capabilities_enabled=true")

		# Test for lockfile if userobject has moved.
		userhasmoved = False
		if new:
			try:
				lockfilename = '/var/lock/deleteuid-%s' % new['uidNumber'][0]
				logger.process('lockfilename: {}'.format(lockfilename))
				lockfiletext = open(lockfilename).read()
				logger.process('lockfiletext: {}'.format(lockfiletext))
				if lockfiletext:
					oxuidnumber = lockfiletext
					userhasmoved = True
					with AsRoot():
						os.remove(lockfilename)
			except:
				pass
		# newcontext != oldcontext ?
		# test, ob benutzer im context schon existiert
		if userhasmoved:
			changeuser = [
				"changeuser",
				"--userid=%s" % oxuidnumber,
				"--username=%s" % username,
			] + authArgs
			logger.info('Modifying OX username {}, oxuid {} in context {}'.format(username, oxuidnumber, contextnum))
			ox_call(changeuser, logger=logger)
			# update OX uid cache
			_ox_user_id_cache.rename_user_by_id(contextnum, username, oxuidnumber)

		if ((contextnum == oldcontextnum) and oxuidnumber) or userhasmoved and oldIsOxUser and newIsOxUser:
			# aendern (6)
			logger.info('aendern (6)')
			if new.get('mailPrimaryAddress'):
				defaultsenderaddress = new['mailPrimaryAddress'][0]
				# Aktuelle default-address auslesen

				oxdefaultsenderaddress = ''
				for row in ox_list_user(admuser, admpass, contextnum, searchpattern=username, ignorecase=True, logger=logger):
					if row[csv_user_row_username] == username:
						oxdefaultsenderaddress = row[csv_user_row_default_sender_address]

				if oxdefaultsenderaddress not in aliaseslist:
					groupwareoptions.append("--defaultsenderaddress=%s" % defaultsenderaddress)

			if new.get('oxLanguage'):
				language = new['oxLanguage'][0]
				oldlanguage = ''
				if old.get('oxLanguage'):
					oldlanguage = old['oxLanguage'][0]
				if language != oldlanguage:
					groupwareoptions.append("--language=%s" % language)

			if new.get('oxTimeZone'):
				timezone = new['oxTimeZone'][0]
				oldtimezone = ''
				if old.get('oxTimeZone'):
					oldtimezone = old['oxTimeZone'][0]
				if timezone != oldtimezone:
					groupwareoptions.append("--timezone=%s" % timezone)

			if email == oldemail:
				# changeuser returns error message if --email is given as argument but value dit not change
				# ===>
				# user user2 in context ?? could not be changed:
				# Server response:
				# 	 Primary mail address already exists in this context.
				newgroupwareoptions = []
				for i in groupwareoptions:
					if not i.startswith("--email="):
						newgroupwareoptions.append(i)
				groupwareoptions = newgroupwareoptions
			cmd = [
				"changeuser",
				"--username=%s" % username,
			] + authArgs + groupwareoptions
			logger.info('Modifying OX settings for uid "{}" in context {}'.format(username, contextnum))
			ox_call(cmd, logger=logger)
			ox_call(["changeuser", "--username=%s" % (username,)] + authArgs + groupware_access_options, logger=logger)
		else:
			# erzeugen (1) (2) (5)
			logger.info('erzeugen (1) (2) (5)')
			if newIsOxUser:
				if not oxuidnumber:
					cmd = ["createuser"] + authArgs
					logger.process('Creating ox user "{}" in context {}'.format(username, contextnum))
				else:
					cmd = ["changeuser"] + authArgs
					logger.info('Modifying OX settings for uid "{}" in context {}'.format(username, contextnum))

				if new.get('mailPrimaryAddress'):
					groupwareoptions.append("--defaultsenderaddress=%s" % new['mailPrimaryAddress'][0])

				if new.get('oxLanguage'):
					groupwareoptions.append("--language=%s" % new['oxLanguage'][0])

				if new.get('oxTimeZone'):
					groupwareoptions.append("--timezone=%s" % new['oxTimeZone'][0])

				if username == admuser:
					password = admpass
				else:
					password = "dummy"

				cmd.append("--username=%s" % username)
				cmd.append("--password=%s" % password)
				cmd = cmd + groupwareoptions + groupware_access_options

				ox_call(cmd, logger=logger)
				ox_call(["changeuser", "--username=%s" % username] + authArgs + groupware_access_options, logger=logger)

				if not oxuidnumber:
					# try to fetch OX user id for new user directly from OX and store it in OXUserIdCache
					oxuidnumber = get_ox_id(admuser, admpass, contextnum, username, fetch_from_ox=True)
				if not oxuidnumber:
					# this should never happen
					logger.error(
						'DB didn\'t return oxuidnumber for username "{username}" in context {context}. Probably '
						'not an OX user.\nnew: {new}\nold: {old}'.format(
							username=username, new=new, old=old, context=contextnum))
					raise Exception("User '{0}' doesn't exist after creation in context {1}.".format(username, contextnum))

		filter_s = filter_format('(&(memberUid=%s)(objectClass=posixGroup))', (username,))
		result_id = ldapCon.search(filter=filter_s, attr=['cn'])
		if result_id:
			# logger.info('groups found')
			oxgroups = [prigroup]
			for w in result_id:
				if w[1].values()[0][0] != prigroup:
					oxgroups.append(w[1].values()[0][0])
			logger.process('oxgroups: {}'.format(oxgroups))

			groupdisplayname = None
			# Benutzer zu Gruppen hinzufügen, bzw. Gruppe mit Benutzer anlegen:
			if oxuidnumber:
				for group in oxgroups:
					# logger.info('processing group {}'.format(w[1].values()[0][0]))
					# Gruppe suchen:
					foundgroup = found_user = False
					for grp in _ox_group_cache.query_one_context(contextnum):
						if grp.groupname.lower() == group.lower():
							foundgroup = True
							if oxuidnumber in grp.member_list:
								found_user = True
								logger.info('User {!r} in context {} already in group {!r}.'.format(
									username, contextnum, grp.groupname))
							break
					if found_user:
						continue

					# Displayname zu Gruppe finden:
					# 'description'
					# 'cn'
					filter_s = filter_format('(&(cn=%s)(objectClass=oxGroup))', (group,))
					result_id = ldapCon.search(filter=filter_s, attr=['cn', 'description'])
					# logger.info('result_id (group): {}({})'.format(result_id, group))
					groupdisplayname = group
					if result_id:
						if not foundgroup:
							logger.process('creating ox group: "{}" in context {}'.format(group, contextnum))
						if result_id[0][1].get('description'):
							groupdisplayname = result_id[0][1]['description'][0]
							# logger.info('groupsdisplayname: {}'.format(groupdisplayname))
						# gruppe existiert nicht und sicherheitshalber, falls die Gruppe doch bereits exisitiert.
						cmd = [
							"changegroup" if foundgroup else "creategroup",
							"--name=%s" % group,
							"--displayname=%s" % groupdisplayname,
							"--addmembers=%s" % oxuidnumber,
						] + authArgs
						logger.info('Handling group membership for uid "{}" in context {}'.format(username, contextnum))
						ox_call(cmd, "Member already exists in group", logger=logger)
						if not foundgroup:
							_ox_group_cache.add_from_ox(contextnum, group, admuser, admpass)
						_ox_group_cache.add_members(contextnum, group, oxuidnumber)
			# Benutzer aus Gruppen entfernen:
			# OX-Gruppen des User-Objektes finden, in denen der Benutzer ist (oxgroups)
			# alle anderen Gruppen auf OX suchen (oxgroupmap)
			nooxgroups = []
			for grp in _ox_group_cache.query_one_context(contextnum):
				if grp.groupname not in oxgroups:
					if grp.groupname != 'users':
						tmpoxgroup = [
							grp.groupname,
							grp.members,
						]
						nooxgroups.append(tmpoxgroup)
			logger.info('User in OX-groups: {}'.format(oxgroups))
			# logger.info('User not in OX-groups: {}'.format(nooxgroups))
			# in diesen Gruppen auf OX suchen, ob Benutzer drin ist:
			for grouprow in nooxgroups:
				groupuserlist = grouprow[1].split(',')
				if oxuidnumber in groupuserlist:
					# ja: Benutzer aus Gruppe entfernen

					# Prüfen, ob Benutzer der letzte der Gruppe war:
					if len(groupuserlist) <= 1:
						# ja: Gruppe löschen
						logger.process('removing ox group: "{}" from context {}'.format(grouprow[0], contextnum))
						deletegroup = [
							"deletegroup",
							"--name=%s" % grouprow[0],
						] + authArgs
						ox_call(deletegroup, "com.openexchange.admin.rmi.exceptions.NoSuchObjectException", logger=logger)
						_ox_group_cache.remove(contextnum, grouprow[0])

					else:
						logger.info('removing ox user from group: "{}" from context {}'.format(grouprow[0], contextnum))
						changegroup = [
							"changegroup",
							"--name=%s" % grouprow[0],
							"--displayname={}".format(groupdisplayname or grouprow[0]),
							"--removemembers=%s" % oxuidnumber,
						] + authArgs
						ox_call(changegroup, "com.openexchange.admin.rmi.exceptions.NoSuchObjectException", logger=logger)
						_ox_group_cache.remove_members(contextnum, grouprow[0], oxuidnumber)

	logger.info('End of ox-user listener handler, running post post_handler hooks...')
	hook_results = run_hook('post_handler_hook', dn, new, old, command, *args, **kwargs)
	logger.info('Hooks returned {!r}'.format(hook_results))
	for result in hook_results:
		if isinstance(result, ContinueHookCommand):
			pass
		elif isinstance(result, ExitHookCommand):
			return result.exit_code
		elif isinstance(result, Exception):
			raise result


def get_context_credentials(contextid):
	admuser = get_context_admin_user(contextid,)
	admpass = get_credentials('/etc/ox-secrets/context%s.secret' % (contextid,))
	return admuser, admpass


def get_ox_id(admuser, admpass, contextnum, username, fetch_from_ox=False):
	'''
	Fetches OX user id for given context and OX username.
	If fetch_from_ox=True, the current value is fetched from OX via "listuser" and stored in the local cache.
	Otherwise the value is fetched from the cache directly.
	'''
	if fetch_from_ox:
		for row in ox_list_user(admuser, admpass, contextnum, searchpattern=username, ignorecase=True, logger=logger):
			if row[csv_user_row_username] == username:
				_ox_user_id_cache.add(contextnum, username, row[csv_user_row_uid])
	return _ox_user_id_cache.query_ox_id(contextnum, username)


def get_cli_option_for_access_flag(name):
	'''
	Returns CLI option name for given access flag. If <name> has not been found in
	mapping table, it is assumed, that the CLI option uses the same spelling.

	>>> get_cli_option_for_access_flag('editpublicfolders')
	'--access-edit-public-folder'
	'''
	return '--access-{}'.format(mapping_module_access_definitions_2_cli.get(name, name))


def get_access_definitions():  # type: () -> Dict[str, Set[str]]
	"""
	Parse ACCESS_DEFINITIONS_FILE for ox access definitions.

	:return: dict: access definition (str) -> capabilities to enable (set)
	"""
	ts = os.stat(ACCESS_DEFINITIONS_FILE).st_mtime  # will raise OSError if file inaccessible
	if not _access_definitions['ts'] or _access_definitions['ts'] < ts:
		_access_definitions['ts'] = ts
		_access_definitions['defs'] = dict(__all=set())
		regex = re.compile(r'^(\w+)=(.+)')
		for line in open(ACCESS_DEFINITIONS_FILE, 'rb'):
			m = regex.match(line)
			if not m:
				continue
			name = m.groups()[0].strip()
			capabilities = {x.strip() for x in m.groups()[1].split(',')}
			_access_definitions['defs'][name] = capabilities
			_access_definitions['defs']['__all'].update(capabilities)
		logger.info('found the following access definitions in "{}":\n{}'.format(
			ACCESS_DEFINITIONS_FILE, _access_definitions))
	return _access_definitions['defs']


def get_access_cmdline_args(access_combination_name, changeuser, mobility, ox_drive_enabled):
	"""
	Create cmdline arguments for changeuser regarding feature access.

	:param access_combination_name: str: UDM oxAccess, currently one of 'none',
	'webmail', 'pim', 'groupware', 'premium'.
	:param changeuser: bool: whether to create arguments for changeuser (True)
	or createuser (False)
	:param mobility: bool: whether to enable mobility access (UDM oxAccessUSM)
	:param ox_drive_enabled: bool: whether to enable OX drive for this user
	:return: list: list of strings
	"""
	access_definitions = get_access_definitions()
	if access_combination_name == 'none':
		res = ['--mailenabled=false']
		disable_caps = access_definitions['__all']
		enable_caps = set()
	else:
		access_definition_name = 'oxseforucs_{}{}{}'.format(
			access_combination_name,
			'_plus' if access_combination_name == 'pim' else '',
			'_usm' if mobility else '')
		res = ['--mailenabled=true']
		try:
			enable_caps = access_definitions[access_definition_name].copy()
			disable_caps = access_definitions['__all'].copy() - access_definitions[access_definition_name]
		except KeyError:
			raise ValueError('Unknown access definition %r.', access_definition_name)

		if not ox_drive_enabled or 'infostore' not in enable_caps:
			# if oxAccess or oxDrive deny the usage of OX Drive then remove the capability
			if 'infostore' in enable_caps:
				enable_caps.remove('infostore')
			disable_caps.add('infostore')
		# the else case is not required, because then enable_caps and disable_caps contain the correct values

	for cap in enable_caps:
		res.append('{}=on'.format(get_cli_option_for_access_flag(cap)))
	for cap in disable_caps:
		res.append('{}=off'.format(get_cli_option_for_access_flag(cap)))
	return res


def get_context_id_from_domain(ldapCon, domainname):
	filter_s = filter_format('(&(objectClass=oxMailDomainServerSettings)(cn=%s))', (domainname,))
	query_result = ldapCon.search(filter=filter_s, attr=['oxContextName'])
	if not query_result:
		return ''
	context_name = query_result[0][1].values()[0][0]

	filter_s = filter_format('(&(objectClass=oxContext)(cn=%s))', (context_name,))
	query_result = ldapCon.search(filter=filter_s, attr=['oxContextIDNum'])
	if not query_result:
		return ''
	return query_result[0][1].values()[0][0]


def get_context_id_from_user(old_or_new):
	return old_or_new.get('oxContextIDNum', [''])[0]


def get_ox_home_server_for_context(ldapCon, context_id):
	result = ldapCon.search(
		filter=filter_format('(&(objectClass=oxContext)(oxContextIDNum=%s))', (str(context_id),)),
		attr=['oxHomeServer'])
	if result:
		return result[0][1]['oxHomeServer'][0]
	else:
		raise RuntimeError('Cannot find context information for ID %r. The context does not exist (anymore) in LDAP.' % (context_id,))


if __name__ == '__main__':
	handler.process_pending()
